extend class StickyBomb
{
	bool sticktomonsters;
	bool chainreaction;
	bool selfdestruct;
	bool priming;
	int lifetime;
	int countdown;
	int warningcountdown;
	int checkradius;
	int blastdmg;
	int directdmg;
	bool checkedsurface;
	double floordiff;
	double ceilingdiff;
	actor victim;
	vector2 victimOfs;
	double victimOfsZ;
	actor obstacle;
	vector2 obstacleOfs;
	double obstacleOfsZ;
	Default
	{
		Radius 4;
		Height 10;
		Speed 30;
		scale 0.6;
		Gravity 0.3;
		Projectile;
		+RANDOMIZE;
		+DONTTHRUST;
		-NOGRAVITY;
		-NOBLOCKMAP;
		+HITTRACER;
		
	}
	States
	{
	Spawn:
		VSB2 ABC 1;
		Loop;
	Death:
		VSB1 # 0
		{
			CheckProximity(checkradius);
			if (priming && victim) A_Startsound("SBLauncher/StickMonster");
			else A_Startsound("SBLauncher/Stick");
		}
	Stick:
		VSB1 # 1 bright
		{
			CheckSurface();
			CheckObstacle();
			CheckProximity(checkradius);
		}
		Loop;	
	DelayExplosion:
		VSB1 # 4;
	Explosion:
		VSBX A 4 bright
		{
			if (random(0,1)) bSPRITEFLIP = true;
			A_Setscale(1.0);
			A_Stop();
			bNOGRAVITY = true;
			A_Startsound("SBLauncher/Explode");
			A_Explode(blastdmg, 128, XF_HURTSOURCE|XF_CIRCULAR, fulldamagedistance:64);
			if (tracer)	tracer.DamageMobj(self, target, directdmg, "none");
		}
		VSBX BCDE 4 bright;
		Stop;
	}
	
	void CheckProximity(double rad) //checks for the player, other bombs and monsters
	{
		actor obj;
		let it = BlockThingsIterator.Create(self, rad);
		while (it.Next())
		{
			obj = it.thing;
			//if the player touches the bomb, turn it into an item that can be picked up
			if (obj.player && obj.CountInv("StickyBombAmmo") < obj.GetAmmoCapacity("StickyBombAmmo") && Distance2D(obj) <= obj.radius + radius && pos.z - floorz <= obj.pos.z - floorz + obj.height)
			{
				A_SpawnItemEx("StickyBombAmmo");
				self.destroy();
				break;
			}
			//explode if near another bomb that explodes
			if (chainreaction && obj != self && obj.GetClass() == "StickyBomb" && Distance3D(obj) <= rad) 
			{
				if (obj.InStateSequence(obj.CurState, obj.ResolveState("Explosion")))
				{
					SetStateLabel("DelayExplosion");
					break;
				}
			}
			//if the proximity check found a monster, stick to it
			if (obj.bISMONSTER && obj.health > 0 && Distance2D(obj) <= obj.radius*1.1 + rad/2)
			{
				tracer = obj;
				if (sticktomonsters)
				{
					if (!victim) 
					{
						victim = obj;
						victimOfsZ = pos.z - victim.pos.z;
						victimOfsZ = clamp(victimOfsZ, victim.height*0.05, victim.height*0.9);
						victimOfs = AngleToVector(victim.angle - victim.angleto(self), victim.radius/1.5);
					}
					break;
				}
				else //or explode
				{			
					SetStateLabel("Explosion");
					break;
				}
			}	
		}
		if (victim && victim.health > 0)
		{
			bNOGRAVITY = true;
			Warp(victim, victimOfs.x, victimOfs.y, victimOfsZ, flags:WARPF_INTERPOLATE);
			if (priming)
			{
				countdown--;
				if (countdown <= 0) SetStateLabel("Explosion");
			}
		}
		else if (victim && victim.health < 1)
		{
			bNOGRAVITY = false;
			victim = null;
		}
		if (selfdestruct)
		{
			lifetime--;
			if (lifetime <= 0) SetStateLabel("Explosion");
		}
	}

	void CheckSurface() //checks for walls and ceilings
	{
		if (floorz == ceilingz) SetStateLabel("Explosion"); //explode when crushed between floor and ceiling
		FLineTraceData surface;
		bool hit = LineTrace(angle, 256, pitch, TRF_SOLIDACTORS, data:surface);
		if (hit && pos.z != floorz)
		{
			if (surface.HitType == TRACE_HitCeiling)
			{
				if (pitch < 0) pitch = -90;
				if (!checkedsurface) //check for blacklisted textures on ceilings
				{
					bNOGRAVITY = true;
					for (int i = 0; i < nonsticky.size(); i++)
					{
						let tex = TexMan.CheckForTexture(nonsticky[i]); 
						if (surface.HitTexture == tex)
						{
							bNOGRAVITY = false;
							checkedsurface = true;	
							break;
						}
					}	
				}
			}
			if (surface.HitType == TRACE_HitWall)
			{	
				if (pitch != 0) pitch = 0; 
				if (!checkedsurface) //check for blacklisted textures on walls
				{
					bNOGRAVITY = true;
					for (int i = 0; i < nonsticky.size(); i++)
					{
						let tex = TexMan.CheckForTexture(nonsticky[i]);
						if (surface.HitTexture == tex)
						{
							bNOGRAVITY = false;
							console.printf("angle %d - %s hit",angle,nonsticky[i]);
							checkedsurface = true;	
							break;
						}
					}	
				}
				//check if the wall is moving up or down and move the bomb along with it
				sector tsec;
				sector bsec = surface.HitLine.backsector;
				sector fsec = surface.HitLine.frontsector;
				vector2 hitpos = surface.HitLocation.xy;
		
				if (bsec) tsec = bsec;
				if (surface.HitSector != fsec) tsec = fsec;
				if (tsec)
				{
					let tfloorZ = tsec.floorplane.ZatPoint(hitpos);
					let tceilingZ = tsec.ceilingplane.ZatPoint(hitpos);
					if (tsec.PlaneMoving(Sector.floor))
					{
						if (!floordiff) floordiff = tfloorZ - pos.z;
						if (pos.z <= floorz) SetZ(floorz);
						else SetZ(tfloorZ - floordiff);
					}
					if (tsec.PlaneMoving(Sector.ceiling))
					{
						if (!ceilingdiff) ceilingdiff = tceilingZ - pos.z;
						if (pos.z >= ceilingz - height) SetZ(ceilingz - height);
						else SetZ(tceilingZ - ceilingdiff);
					}
				}
			}
		}
		if (!hit) bNOGRAVITY = false;
	}
	
	void CheckObstacle() //checks for obstacles and decorations
	{
		FLineTraceData obst;
		bool hit = LineTrace(angle, 40, pitch, TRF_SOLIDACTORS, data:obst);
		if (hit && obst.HitType == TRACE_HitActor && !obst.HitActor.bISMONSTER && !obst.HitActor.player && obst.HitType != TRACE_HitWall && obst.HitType != TRACE_HitCeiling)
		{
			obstacle = obst.HitActor;
			obstacleOfsZ = pos.z - obstacle.pos.z;
			obstacleOfsZ = clamp(obstacleOfsZ, obstacle.height*0.05, obstacle.height*0.9);
			obstacleOfs = AngleToVector(obstacle.angle - obstacle.angleto(self), obstacle.radius);	
			bNOGRAVITY = true;
		}
		if (obstacle) Warp(obstacle, obstacleOfs.x, obstacleOfs.y, obstacleOfsZ,flags:WARPF_INTERPOLATE);
		if (!hit) bNOGRAVITY = false;
	}
}